#ifndef _PORT_H
#define _PORT_H

#include "inet.h"

#ifndef NOEXTERNS
extern short		port_inuse (struct in_data *, unsigned short port);
extern struct in_data	*port_find (struct in_data *, unsigned short port);
extern unsigned short	port_alloc (struct in_data *);
#endif

#endif /* _PORT_H */
