/*
 *	Loadable MintNet network interface startup code.
 *
 *	06/22/94, Kay Roemer.
 */

#include "config.h"
#include "netinfo.h"
#include "file.h"
#include "kerbind.h"

extern long driver_init (void);

struct netinfo *NETINFO;
struct kerinfo *KERINFO;

long
init (ker, net)
	struct kerinfo *ker;
	struct netinfo *net;
{
	KERINFO = ker;
	NETINFO = net;

	return (driver_init () != 0) ? 1 : 0;
}
