/*
 *	File for keeping version information.
 *
 *	11/12/93, kay roemer.
 */

#ifndef _VERSION_H
#define _VERSION_H

#define NET_MAJ		0
#define NET_MIN		6
#define NET_PL		0
#define NET_STATUS	"alpha"

#define stringify(x)	_stringify(x)
#define _stringify(x)	#x

#endif /* _VERSION_H */
