/*
 *	Signal related definitions.
 *
 *	11/12/93, kay roemer
 */

#ifndef _SIGNAL_H
#define _SIGNAL_H

#define SIGPIPE		13	/* broken socket connection */	
#define SIGURG		16	/* urgent condition on IO channel */
#define SIGIO		23	/* IO possible */

#endif /* _SIGNAL_H */
