#ifndef _PL0_H
#define _PL0_H

extern void	pl0_set_strobe		(short);
extern void	pl0_set_direction	(short);
extern long	pl0_send_pkt		(short, char *, long);
extern long	pl0_recv_pkt		(short, char *, long);
extern char	pl0_got_ack		(void);
extern void	pl0_send_ack		(void);
extern void	pl0_cli			(void);
extern void	pl0_sti			(void);
extern short	pl0_get_busy		(void);

extern short	pl0_init		(void);
extern void	pl0_open		(void);
extern void	pl0_close		(void);

extern void	pl0_busy_int		(void);

extern long	pl0_old_busy_int;

#endif /* _PL0_H */
