#ifndef _UDP_H
#define _UDP_H

/* Must be enough for IP and ICMP header */
#define UDP_RESERVE	140

/* UDP datagram header */
struct udp_dgram {
	unsigned short	srcport;	/* source port */
	unsigned short	dstport;	/* destination port */
	unsigned short	length;		/* total dgram length */
	unsigned short	chksum;		/* optional checksum */
	char		data[0];	/* begin of data */
};

#define UDP_MINLEN	(sizeof (struct udp_dgram))

#ifndef NOEXTERNS
extern void	udp_init (void);
#endif

#endif /* _UDP_H */
