#ifndef _DUMMYDEV_H
#define _DUMMYDEV_H

#include "file.h"

extern long	dummydev_open	(FILEPTR *);
extern long	dummydev_write	(FILEPTR *, char *, long);
extern long	dummydev_lseek	(FILEPTR *, long, short);
extern long	dummydev_ioctl	(FILEPTR *, short, void *);
extern long	dummydev_datime	(FILEPTR *, short *, short);
extern long	dummydev_close	(FILEPTR *, short);
extern long	dummydev_select	(FILEPTR *, long, short);
extern void	dummydev_unselect(FILEPTR *, long, short);

extern long	dummydev_init	(char *, struct dev_descr *);

#endif /* _DUMMYDEV_H */

